#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>

#include <gmath/gmath.h>

#include "global.h"

/* TODO: fix those */
#include "camera.h"
#include "mesh.h"
#include "object.h"
#include "renderer.h"
#include "scene.h"
#include "shader_manager.h"

#include "opengl/opengl.h"
#include "vulkan/vk.h"

/* static functions */

static bool init();
static void cleanup();
static void display();

/* glfw callbacks */

static void clbk_key(GLFWwindow *win, int key, int scancode, int action, int mods);
static void clbk_motion(GLFWwindow *win, double x, double y);
static void clbk_mouse(GLFWwindow *win, int button, int action, int mods);
static void clbk_reshape(GLFWwindow *win, int width, int height);

/* global variables */

bool use_vulkan = false;
Mat4 mprojection;

GLFWwindow *win;
int win_w = 800;
int win_h = 600;

float phi = 25;
float theta = 0;
float dist = 4;

ShaderManager *sdr_man;

/* variables */
static float aspect;
static Scene *scene;
static OrbitCamera *camera;
static Renderer *rdefault; // default renderer

int main(int argc, char **argv)
{
	for(int i=0; i<argc; ++i) {
		if(strcmp(argv[i], "-opengl") == 0) {
			printf("Backend: OpenGL.\n");
		}
		else if(strcmp(argv[i], "-vulkan") == 0) {
			use_vulkan = true;
			printf("Backend: Vulkan.\n");
		}
		else {
			printf("No backend specified. Using OpenGL.\n");
		}
	}

	if(!init()) {
		fprintf(stderr, "Failed to initialize program.\n");
		return 1;
	}

	glfwSetKeyCallback(win, clbk_key);
	glfwSetCursorPosCallback(win, clbk_motion);
	glfwSetMouseButtonCallback(win, clbk_mouse);
	glfwSetWindowSizeCallback(win, clbk_reshape);

	glfwGetWindowSize(win, &win_w, &win_h);
	clbk_reshape(win, win_w, win_h);

	while(!glfwWindowShouldClose(win)) {
		display();

		glfwSwapBuffers(win);
		glfwPollEvents();
	}

	cleanup();
	// atexit(cleanup);
	return 0;
}

static bool init()
{
	if(use_vulkan) {
		if(!init_vulkan())
			return false;
	}
	else {
		if(!init_opengl())
			return false;
	}

	sdr_man = new ShaderManager;

	camera = new OrbitCamera;
	camera->set_orbit_params(phi, theta, dist);

	scene = new Scene;
	if(!scene->load("data/spot/spot.obj")) {
		fprintf(stderr, "Failed to load scene.\n");
		return false;
	}

	rdefault = new Renderer;
	rdefault->camera = camera;
	rdefault->scene = scene;

	if(!rdefault->create()) {
		fprintf(stderr, "Failed to create default renderer.\n");
		return false;
	}

// TODO delete: debugging
	for(size_t i=0; i<scene->objects.size(); ++i) {
		printf("object: %d\n", (int)i);
		printf("mesh: %s\n", scene->objects[i]->mesh->name.c_str());
		printf("material: %s\n", scene->objects[i]->material->name.c_str());
		printf("transform:\n");
		scene->objects[i]->transform.print();
	}
	return true;
}

static void cleanup()
{
	delete sdr_man;
	delete camera;
	delete rdefault;

	if(use_vulkan) {
		cleanup_vulkan();
	}
	else {
		cleanup_opengl();
	}
}

static void clbk_key(GLFWwindow *win, int key, int scancode, int action, int mods)
{
	if(key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) {
		glfwSetWindowShouldClose(win, GLFW_TRUE);
	}
}

static double prev_x, prev_y;
static bool button[8];

static void clbk_motion(GLFWwindow *win, double x, double y)
{
	double dx = x - prev_x;
	double dy = y - prev_y;

	prev_x = x;
	prev_y = y;

	if(button[0]) {
		theta += dx;
		phi += dy;

		if(phi < -90)
			phi = -90;
		if(phi > 90)
			phi = 90;
	}

	if(button[1]) {
		dist += dy;
		if(dist < 0.0) {
			dist = 0.0;
		}
	}
}

static void clbk_mouse(GLFWwindow *win, int bn, int action, int mods)
{
	button[bn] = action == GLFW_PRESS;
	glfwGetCursorPos(win, &prev_x, &prev_y);
}

static void clbk_reshape(GLFWwindow *win, int width, int height)
{
	if(use_vulkan) {
		//TODO
		return;
	}
	else {
		glViewport(0, 0, width, height);
		aspect = (float)width / (float)height;
		mprojection = calc_projection_matrix(45, aspect, 0.5, 1000.0);
	}

	win_h = height;
	win_w = width;
}

static void display()
{
	camera->set_orbit_params(phi, theta, dist);
	rdefault->draw();
}