#include <GL/glew.h>

#include "global.h"

#include "camera.h"
#include "mesh.h"
#include "object.h"
#include "renderer.h"
#include "scene.h"
#include "shader.h"
#include "shader_manager.h"
#include "texture.h"

Renderer::Renderer()
{
	scene = 0;
	camera = 0;
	sprog = 0;
}

Renderer::~Renderer()
{
}

bool Renderer::create()
{
	if(!(sprog = sdr_man->create_shader_program("default.v.glsl", "default.f.glsl"))) {
		return false;
	}

	return true;
}

void Renderer::draw() const
{
	if(!camera || !scene)
		return;

	glClearColor(0.5, 0.5, 0.5, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if(!sprog->link())
		return;

	sprog->use();

	for(size_t i=0; i<scene->objects.size(); i++) {
		draw_object(scene->objects[i]);
	}
}

void Renderer::draw_object(Object *object) const
{
	Material *m = object->material;
	/* diffuse */
	int diff_loc;
	if((diff_loc = sprog->get_uniform_location("diffuse")) != -1) {
		sprog->set_uniformf(diff_loc, m->diffuse.x, m->diffuse.y, m->diffuse.z, 1);
	}
	/* specular */
	int spec_loc;
	if((spec_loc = sprog->get_uniform_location("specular")) != -1) {
		sprog->set_uniformf(spec_loc, m->specular.x, m->specular.y, m->specular.z, 1);
	}
	/* specular exponent */
	int shin_loc;
	if((shin_loc = sprog->get_uniform_location("shininess")) != -1) {
		sprog->set_uniformf(shin_loc, m->shininess);
	}
	/* texture */
	if (m->dtex)
		m->dtex->bind();

	// Mat4 mmviewproj = mprojection * camera->get_view_matrix() * object->transform;
	Mat4 mmviewproj = object->transform * camera->get_view_matrix() * mprojection;
	// printf("\nmatrices:\n");
	// mprojection.print();
	// camera->get_view_matrix().print();
	// object->transform.print();
	// mmodelview.print();

	int mmviewproj_loc;
	if((mmviewproj_loc = sprog->get_uniform_location("mmviewproj")) != -1) {
		sprog->set_uniform_matrix(mmviewproj_loc, mmviewproj);
	}

	Mat4 mview = camera->get_view_matrix(); //* object->transform;
	int mview_loc;
	if((mview_loc = sprog->get_uniform_location("mview")) != -1) {
		sprog->set_uniform_matrix(mview_loc, mview);
	}

	object->mesh->update_vertex_data();
	object->mesh->draw();
}